-------------------------------------------------------------------------------
--                G  O  B  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

-- Labels for the key bindings
_G["BINDING_HEADER_NIGGLES"]                     = L["Niggles"];
_G["BINDING_NAME_CLICK NigglesMount:LeftButton"] = L["Mount"];
_G["BINDING_NAME_NIGGLES_TOGGLELOCATOR"]         = L["Toggle Locator"];

-- Account & Character data saved for the Niggles addon
NigglesDB  = {};
NigglesDBC = {};

-- Shapeshift forms for classes that should ALWAYS have an active form
local classShapeshiftForm =
{
  ["DEATHKNIGHT"] = L["Presence"],
  ["HUNTER"]      = L["Aspect"],
  ["MONK"]        = L["Stance"],
  ["WARRIOR"]     = L["Stance"],
};

-------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local HELP_COLOR_CODE = "|cFF8080FF";

-- Required Account and Character data saved for the Niggles addon
local requiredDB =
{
  ["raidAssists" ] = 1,
  ["moderators"  ] = 1,
  ["nonRaiders"  ] = 1
};
local requiredDBC =
{
  ["faction"     ] = 1,
  ["settings"    ] = 1,
  ["mounts"      ] = 1,
  ["pets"        ] = 1,
  ["achievements"] = 1
};

-- Default values for all Niggles Character settings
local defaultSettings =
{
  -- General Options
  ["UiAchievementSearch"]  = false,
  ["UiAutoLootArch"]       = false,
  ["UiAutoLootFish"]       = false,
  ["UiAutoLootMoney"]      = false,
  ["UiChatClassColor"]     = true,
  ["UiEMIgnoreToggle"]     = true,
  ["UiFirstLastPage"]      = true,
  ["UiFlightList"]         = true,
  ["UiGuildRosterFilter"]  = true,
  ["UiItemDelete"]         = true,
  ["UiMailFont"]           = true,
  ["UiMapLocation"]        = true,
  ["UiMerchantBuyMax"]     = true,
  ["UiMerchantSearch"]     = true,
  ["UiMountsFilter"]       = true,
  ["UiMiniMapZoom"]        = true,
  ["UiQuestRewards"]       = true,

  -- Warning Options
  ["WarningCurrency"]      = true,
  ["WarningCurrencyMask"]  = 0x02,
  ["WarningEquipSetItem"]  = true,
  ["WarningRod"]           = true,
  ["WarningLance"]         = true,
  ["WarningLowLevelItems"] = true,
  ["WarningNoForm"]        = true,
  ["WarningPetBattle"]     = true,
  ["WarningPetGrowl"]      = false,
  ["LogChat"]              = false,
  ["LogCombat"]            = false,

  -- Profession Options
  ["ProfArchHilite"]       = true,
  ["ProfArchKeystones"]    = true,
  ["ProfArchEnhIcon"]      = true,
  ["ProfArchRClickBack"]   = true,
  ["ProfArchProgress"]     = true,

  -- Raids Options
  ["RaidsTarget"]          = true,
  ["RaidsAutoPromote"]     = true,
  ["RaidsAutoLootMethod"]  = true,
  ["RaidsLootMethod"]      = "master",
  ["RaidsAutoLootThold"]   = true,
  ["RaidsLootThreshold"]   = 4,

  -- Calendar Event Options
  ["CalEventsList"]        = true,
  ["CalEventsInvites"]     = true,
  ["CalEventsOnline"]      = true,
  ["CalEventsFilter"]      = false,
  ["CalEventsInactive"]    = 60,

  -- Companion Options
  ["MountsEnabled"]        = true,
  ["MountsClassRace"]      = true,
  ["MountsHoliday"]        = false,
  ["PetsEnabled"]          = true,
  ["PetsInEKingdoms"]      = true,
  ["PetsInKalimdor"]       = true,
  ["PetsInNorthrend"]      = true,
  ["PetsInOutland"]        = true,
  ["PetsInPandaria"]       = true,
  ["PetsInCities"]         = true,
  ["PetsInScenarios"]      = true,
  ["PetsInDungeons"]       = true,
  ["PetsInRaids"]          = false,
  ["PetsInArenas"]         = false,
  ["PetsInBG"]             = false,

  -- Locator Options
  ["LocatorShown"]         = false
};

-- Valid commands for Niggles
local validCommands =
{
  {command = L["General"]:lower(),     category = "optionsGeneral"    },
  {command = L["Warnings"]:lower(),    category = "optionsWarnings"   },
  {command = L["Professions"]:lower(), category = "optionsProfessions"},
  {command = L["Raids"]:lower(),       category = "optionsRaids"      },
  {command = L["Events"]:lower(),      category = "optionsCalEvents"  },
  {command = L["Companions"]:lower(),  category = "optionsCompanions" }
};

-------------------------------------------------------------------------------
--                   S  L  A  S  H     C  O  M  M  A  N  D  S
-------------------------------------------------------------------------------

SLASH_NIGGLES1 = "/niggles";
SLASH_NIGGLES2 = "/ngl";

-------------------------------------------------------------------------------
--                 L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to make a simple copy of a table
--
local function tableCopy(table)
  local copy = {};
  for key, value in pairs(table) do
    copy[key] = value;
  end
  return copy;
end

-------------------------------------------------------------------------------
--                G  L  O  B  A  L    F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to process the slash command
--
function SlashCmdList.NIGGLES(message, editbox)
  -- Local Variables
  local command, arguments = message:match("^(%S*)%s*(.-)$");
  local category = nil

  -- Check a command was specified
  if ((command ~= nil) and (command ~= "")) then
    -- Search the valid commands for one that matches the specified command
    for _, cmdInfo in ipairs(validCommands) do
      if (cmdInfo.command:find("^"..command)) then
        category = cmdInfo.category;
        break;
      end
    end
  end

  -- Check if the command was found
  if (category ~= nil) then
    InterfaceOptionsFrame_OpenToCategory(Niggles[category]);
  else
    -- Display the help
    DEFAULT_CHAT_FRAME:AddMessage(HELP_COLOR_CODE.."/niggles||/ngl");
    for _, cmdInfo in ipairs(validCommands) do
      DEFAULT_CHAT_FRAME:AddMessage(HELP_COLOR_CODE..
        string.gsub(cmdInfo.command, "(.)(.*)", "  %1[%2]"));
    end
  end

  return;
end

--
-- Function to process events
--
function NigglesOnEvent(self, event, ...)
  -- Process the event that occurred
  if (event == "UNIT_SPELLCAST_SUCCEEDED") then
    -- Check the spell is 'Growl' from a friendly pet in a multi-player party
    local unitId, _, _, _, spellId = ...;
    if ((spellId == 2649) and 
        (UnitIsFriend("player", unitId)) and
        (GetNumGroupMembers() > 1)) then
      -- Check the player is in a PvE instance
      local _, instanceTye = IsInInstance();
      if ((instanceType == "dungeon") or
          (instanceType == "raid")) then
        Niggles:AddWarning(format(L["Pet Growled"], UnitName(unitId)));
        PlaySound("RaidWarning");
      end
    end
  elseif (event == "PLAYER_REGEN_DISABLED") then
    -- Check if the player is in a vehicle
    if (not UnitInVehicle("player")) then
      -- Local Variables
      local playSound = false;
      local numWarnings = 0;

      -- Get the ID of the weapon in the player's main hand
      local itemId = GetInventoryItemID("player", self.mainHandSlotId);
      if (itemId) then
        -- Display a warning if a lance or fishing rod is equipped
        if ((NigglesDBC.settings.WarningLance) and
            (NigglesEquipmentIsLance(itemId))) then
          Niggles:AddWarning(L["Lance Equipped"]);
          numWarnings = numWarnings+1;
        elseif ((NigglesDBC.settings.WarningRod) and
                (IsEquippedItemType(L["Fishing Pole"]))) then
          Niggles:AddWarning(L["Fishing Pole Equipped"]);
          numWarnings = numWarnings+1;
        end
      end

      -- Display a warning if there are low level items equipped or empty slots
      if (NigglesDBC.settings.WarningLowLevelItems) then
        if (NigglesEquipmentGetLowIlvl()-numWarnings > 0) then
          Niggles:AddWarning(L["Low Level Items Equipped"]);
          numWarnings = numWarnings+1;
        elseif ((NigglesEquipmentGetEmptySlots() > 0) and
                (UnitLevel("player") >= 60)) then
          Niggles:AddWarning(L["Empty Equipment Slots"]);
          numWarnings = numWarnings+1;
        end
      end

      -- Display a warning if there is no active form
      if ((NigglesDBC.settings.WarningNoForm) and
          (GetNumShapeshiftForms() > 0) and
          (GetShapeshiftForm() == 0)) then
        Niggles:AddWarning(format(L["No Form"], Niggles.playerForm));
        numWarnings = numWarnings+1;
      end

      -- Play a sound to alert the user of any messages
      if (numWarnings > 0) then
        PlaySound("RaidWarning");
      end
    end
  elseif (event == "PLAYER_REGEN_ENABLED") then
    if (Niggles.updateRaid == true) then
      Niggles.updateRaid = false;
      NigglesRaidButtonsSetTargets();
    end
  elseif (event == "GROUP_ROSTER_UPDATE") then
    NigglesRaidButtonsSetTargets();
    if ((NigglesDBC.settings.RaidsAutoPromote == true) and
        (UnitIsGroupLeader("PLAYER"))) then
      NigglesRaidRosterPromoteAssists();
    end
    -- TODO: Remove when 'PARTY_CONVERTED_TO_RAID' triggers again
    if ((IsInGroup()) and
        (IsInRaid()) and
        (Niggles.inGroup) and
        (not Niggles.inRaid) and
        (UnitIsGroupLeader("PLAYER"))) then
      if (NigglesDBC.settings.RaidsAutoLootMethod == true) then
        NigglesTimerCreate("RaidsAutoLootMethod", SetLootMethod,
          NigglesDBC.settings.RaidsLootMethod, "player", 1.0, false);
      end
      if (NigglesDBC.settings.RaidsAutoLootThold == true) then
        NigglesTimerCreate("RaidsAutoLootThold", SetLootThreshold,
          NigglesDBC.settings.RaidsLootThreshold, nil, 2.0, false);
      end
    end
    Niggles.inGroup = IsInGroup();
    Niggles.inRaid  = IsInRaid();
  elseif (event == "PARTY_CONVERTED_TO_RAID") then
    if (UnitIsGroupLeader("PLAYER")) then
      if (NigglesDBC.settings.RaidsAutoLootMethod == true) then
        NigglesTimerCreate("RaidsAutoLootMethod", SetLootMethod,
          NigglesDBC.settings.RaidsLootMethod, "player", 1.0, false);
      end
      if (NigglesDBC.settings.RaidsAutoLootThold == true) then
        NigglesTimerCreate("RaidsAutoLootThold", SetLootThreshold,
          NigglesDBC.settings.RaidsLootThreshold, nil, 2.0, false);
      end
    end
  elseif (event == "BAG_UPDATE") then
    local bagId = select(1, ...);
    if ((bagId >=0) and (bagId <= NUM_BAG_SLOTS)) then
      NigglesPetsBagUpdate();
    end
  elseif (event == "UPDATE_STEALTH") then
    -- Check if the player is stealthed
    if (IsStealthed() == 1) then
      -- Dismiss the pet by summoning it again
      local petId = C_PetJournal.GetSummonedPetGUID();
      if (petId ~= nil) then
        C_PetJournal.SummonPetByGUID(petId);
      end
    end
  elseif ((event == "PET_JOURNAL_LIST_UPDATE") or
          (event == "PET_JOURNAL_PET_DELETED") or
          (event == "COMPANION_UNLEARNED")) then
    NigglesPetsUpdate();
  elseif ((event == "QUEST_DETAIL") or
          (event == "QUEST_PROGRESS") or
          (event == "QUEST_COMPLETE")) then
    NigglesQuestsItemInfoUpdate();
  elseif ((event == "SPELLS_CHANGED") or
          (event == "LEARNED_SPELL_IN_TAB")) then
    NigglesMountsSpellsUpdate();
  elseif (event == "GET_ITEM_INFO_RECEIVED") then
    if ((NigglesDBC.settings.UiQuestRewards) and
        ((QuestFrame:IsShown()) or
         (QuestLogFrame:IsShown()) or
         (QuestLogDetailFrame:IsShown()))) then
      NigglesQuestsItemInfoUpdate();
    end
  elseif (event == "GUILD_ROSTER_UPDATE") then
    NigglesCalendarEventsGuildRosterUpdate();
  elseif (event == "PLAYER_EQUIPMENT_CHANGED") then
    NigglesEquipmentScan();
  elseif (event == "CALENDAR_UPDATE_INVITE_LIST") then
    NigglesCalEventInviteListUpdate();
  elseif (event == "CURRENCY_DISPLAY_UPDATE") then
    NigglesCurrencyCheck();
  elseif (event == "ACHIEVEMENT_EARNED") then
    NigglesAchievementEarned(select(1, ...));
  elseif (event == "PET_BATTLE_OPENING_DONE") then
    NigglesEquipmentCheckPetBattle();
  elseif (event == "ADDON_LOADED") then
    local addonName = select(1, ...);
    if (addonName == "Niggles") then
      NigglesOnLoaded();
    elseif (addonName == "Blizzard_AchievementUI") then
      NigglesAchievementsEnable(false);
    elseif (addonName == "Blizzard_ArchaeologyUI") then
      NigglesArchaeologyInitUI();
    elseif (addonName == "Blizzard_Calendar") then
      NigglesCalendarInit();
    elseif (addonName == "Blizzard_GuildUI") then
      NigglesGuildRosterInit();
    elseif (addonName == "Blizzard_PetJournal") then
      NigglesCompanionsInitUI();
    end
  elseif (event == "PLAYER_LOGIN") then
    -- Initialise the Archaeology functionality
    NigglesArchaeologyInitialise();

    -- Initialise the holiday functionality
    NigglesHolidaysInitialise();

    -- Initialise the Mount and Pet functionality
    NigglesCompanionsInit();
    
    -- Initialise the equipment functionality
    NigglesEquipmentInit();

    -- Initialise the changes to the Quest panels
    NigglesQuestsInit();

    -- Show/Hide the Locator frame
    NigglesLocatorShow(NigglesDBC.settings.LocatorShown);

    -- Register for events
    NigglesRegisterForEvents();

    -- Check if the player is in a group and/or raid
    Niggles.inGroup = IsInGroup();
    Niggles.inRaid  = IsInRaid();
  elseif (event == "PLAYER_LOGOUT") then
    -- Shut down the Achievements functionality
    NigglesAchievementsShutdown();
  end

  return;
end

--
-- Function to initialise the addon
--
function NigglesOnLoad(self)
  -- Initialise some variables
  self.updatedOptions = nil;
  self.petMask        = 0;
  self.mainHandSlotId = GetInventorySlotInfo("MainHandSlot");
  self.updateRaid     = false;
  self.timers         =
  {
    elapsed = 0,
    list = nil
  };

  -- Get the player's class, race and form
  _, self.playerClass = UnitClass("player");
  _, self.playerRace  = UnitRace("player");
  self.playerForm = classShapeshiftForm[self.playerClass];

  -- Register the event triggered by the addon being loaded
  self:RegisterEvent("ADDON_LOADED");

  -- Add a method to display warnings
  function self:AddWarning(message)
    self:AddMessage(message, 1.0, 1.0, 0.0, 1.0);
    return;
  end

  return;
end

--
-- Function called once the addon has been loaded
--
function NigglesOnLoaded()
  -- Initialise the settings
  NigglesSettingsInit();

  -- Hook the standard UI
  NigglesHookUi();

  -- Register for some events
  Niggles:RegisterEvent("PLAYER_LOGIN");
  Niggles:RegisterEvent("PLAYER_LOGOUT");
  Niggles:RegisterEvent("SPELLS_CHANGED");
  Niggles:RegisterEvent("LEARNED_SPELL_IN_TAB");
  Niggles:RegisterEvent("GET_ITEM_INFO_RECEIVED");

  return;
end

--
-- Function to cancel changes to the options
--
function NigglesOptionsCancel()
  -- Destroy any updated options
  Niggles.updatedOptions = nil;

  return;
end

--
-- Function to initialise check buttons in the Niggles Options frame
--
function NigglesOptionsCheckButtonInit(checkButton, setting, text, tooltip)
  -- Local Variables
  local textWidth;
  local textFont;

  -- Check if no setting has been specified
  if (setting == nil) then
    -- Work out the button's setting from it's name
    setting = string.gsub(checkButton:GetName(),
      "^"..checkButton:GetParent():GetName().."(.*)$", "%1");
  end

  -- Save the setting the button is for
  checkButton.nglSetting = setting;

  -- Set the check button's font by simulating an enable event
  NigglesOptionsCheckButtonOnEnable(checkButton)

  -- Set the check button's text
  if (text ~= nil) then
    checkButton.text:SetText(text);
  else
    checkButton.text:SetText(L[setting]);
  end

  -- Update the size of the button's hit box
  textWidth = checkButton.text:GetStringWidth();
  if ((textWidth ~= nil) and (textWidth > 0)) then
    checkButton:SetHitRectInsets(0, -textWidth, 0, 0);
  end
 
  -- Set the check button's text
  if (tooltip ~= nil) then
    checkButton.tooltip = tooltip;
  end

  return;
end

--
-- Function to process check button clicks in the Niggles Options frame
--
function NigglesOptionsCheckButtonOnClick(checkButton)
  if (checkButton:GetChecked()) then
    PlaySound("igMainMenuOptionCheckBoxOn");
    Niggles.updatedOptions.settings[checkButton.nglSetting] = true;
  else
    PlaySound("igMainMenuOptionCheckBoxOff");
    Niggles.updatedOptions.settings[checkButton.nglSetting] = false;
  end
end

--
-- Function to process enabling/disabling of a check button
--
function NigglesOptionsCheckButtonOnEnable(checkButton)
  -- Local Variables
  local font = "GameFontHighlight";

  -- Work out which font to use
  if (checkButton:IsEnabled()) then
    if (checkButton.normalFont ~= nil) then
      font = checkButton.normalFont;
    end
  else
    if (checkButton.disabledFont ~= nil) then
      font = checkButton.disabledFont;
    else
      font = "GameFontDisable";
    end
  end

  -- Change the check button's font
  checkButton.text:SetFontObject(font);

  return;
end

--
-- Function to set the tooltip for a widget in the Niggles Options panels
--
function NigglesOptionsCheckButtonSetTooltip(checkButton)
  -- Local Variables
  local tooltipText = nil;

  -- Find any tooltip for the check button
  if (checkButton.tooltip ~= nil) then
    tooltipText = checkButton.tooltip;
  elseif (L[checkButton.nglSetting.."Tooltip"] ~= nil) then
    tooltipText = L[checkButton.nglSetting.."Tooltip"];
  end

  -- Check if a tooltip was found
  if (tooltipText ~= nil) then
    GameTooltip:SetOwner(checkButton, "ANCHOR_TOPLEFT", 20);
    GameTooltip:SetText(tooltipText, nil, nil, nil, nil, 1);
    GameTooltip:Show();
  end

  return;
end

--
-- Function to change all the settings on the Niggles Options panels to their
-- default values.
--
function NigglesOptionsDefaults(frame)
  -- Destroy any updated options
  Niggles.updatedOptions = nil;

  -- Set all the settings to their default values
  for key, value in pairs(defaultSettings) do
    NigglesDBC.settings[key] = value;
  end

  -- Clear the character lists
  NigglesDB.raidAssists = {};
  NigglesDB.moderators  = {};
  NigglesDB.nonRaiders  = {};

  -- Clear all the companions
  NigglesDBC.mounts = {};
  NigglesDBC.pets   = {};

  -- Re-initialise the settings
  NigglesOptionsRefresh();

  return;
end

--
-- Function to update the value of a drop down when one of its options is
-- clicked.
--
function NigglesOptionsDropDownOnClick(self, arg1, arg2, checked)
  arg1:SetValue(self.value);
end

--
-- Function to set the value of a drop down
--
function NigglesOptionsDropDownSetValue(self, value)
  -- Save the new value
  Niggles.updatedOptions.settings[self.nglSetting] = value;

  -- Set the drop down menu's value
  UIDropDownMenu_Initialize(self,
    NigglesOptionsRaidsLootMenuInitialise);
  UIDropDownMenu_SetSelectedValue(self, value);

  return;
end

--
-- Function to process edit box text changes in the Niggles Options frame
--
function NigglesOptionsEditBoxOnTextChange(editBox)
  if (editBox.nglSetting ~= nil) then
    if (editBox:IsNumeric()) then
      Niggles.updatedOptions.settings[editBox.nglSetting] =
        tonumber(editBox:GetText());
    else
      Niggles.updatedOptions.settings[editBox.nglSetting] = editBox:GetText();
    end
  end
end

--
-- Function to refresh the options for the Niggles Options panels
--
function NigglesOptionsRefresh()
  -- Check there aren't any updated options already
  if (Niggles.updatedOptions == nil) then
    -- Initialise the updated options
    Niggles.updatedOptions = {};

    -- Copy the existing settings
    Niggles.updatedOptions.settings = tableCopy(NigglesDBC.settings);

    -- Copy the existing character lists
    Niggles.updatedOptions.raidAssists = tableCopy(NigglesDB.raidAssists);
    Niggles.updatedOptions.moderators  = tableCopy(NigglesDB.moderators);
    Niggles.updatedOptions.nonRaiders  = tableCopy(NigglesDB.nonRaiders);

    -- Copy the existing lists of companions
    Niggles.updatedOptions.mounts = tableCopy(NigglesDBC.mounts);
    Niggles.updatedOptions.pets   = tableCopy(NigglesDBC.pets);

    -- Initialise all the widgets to the correct state for the settings
    NigglesOptionsRefreshSettings(NigglesOptionsGeneral);
    NigglesOptionsRefreshSettings(NigglesOptionsWarnings);
    NigglesOptionsWarningsEnableFrames();
    NigglesOptionsRefreshSettings(NigglesOptionsProfessions);
    NigglesOptionsRefreshRaids();
    NigglesOptionsRefreshCalEvents();
    NigglesOptionsRefreshCompanions();
  end

  return;
end

--
-- Function to refresh settings in Niggles Options panels. This function
-- is recursive.
--
function NigglesOptionsRefreshSettings(widget)
  -- Check if the widget has a Niggles setting associated with it
  if ((widget.nglSetting) and
      (Niggles.updatedOptions.settings[widget.nglSetting] ~= nil)) then
    -- Set the widget's state, based on its type
    local widgetType = widget:GetObjectType()
    if (widgetType == "CheckButton") then
      widget:SetChecked(Niggles.updatedOptions.settings[widget.nglSetting]);
    elseif (widgetType == "EditBox") then
      if (widget:IsNumeric()) then
        widget:SetNumber(Niggles.updatedOptions.settings[widget.nglSetting]);
      else
        widget:SetText(Niggles.updatedOptions.settings[widget.nglSetting]);
      end
      widget:SetCursorPosition(0);
    else
      if (widget.nglObjectType == "UIDropDownMenu") then
        widget:SetValue(Niggles.updatedOptions.settings[widget.nglSetting]);
      end
    end
  end

  -- Process all children of the current widget
  for _, child in ipairs({widget:GetChildren()}) do
    NigglesOptionsRefreshSettings(child);
  end
end

--
-- Function to save updated options for the Niggles Options panels
--
function NigglesOptionsSave()
  -- Check if there are any updated options
  if (Niggles.updatedOptions ~= nil) then
    -- Save the new settings
    NigglesDBC.settings = Niggles.updatedOptions.settings;

    -- Save the new lists of companions
    NigglesDBC.mounts = Niggles.updatedOptions.mounts;
    NigglesDBC.pets   = Niggles.updatedOptions.pets;

    -- Save the new character lists
    NigglesDB.raidAssists = Niggles.updatedOptions.raidAssists;
    NigglesDB.moderators  = Niggles.updatedOptions.moderators;
    NigglesDB.nonRaiders  = Niggles.updatedOptions.nonRaiders;

    -- Register for events
    NigglesRegisterForEvents();

    -- Remove the updated options
    Niggles.updatedOptions = nil;

    -- Force any visible calendar event frame to update itself
    if ((CalendarCreateEventFrame) and
        (CalendarCreateEventFrame:IsShown())) then
      NigglesCalenderInvitesShow(CalendarCreateEventFrame);
    end
    if ((CalendarViewEventFrame) and
        (CalendarViewEventFrame:IsShown())) then
      NigglesCalenderInvitesShow(CalendarViewEventFrame);
    end

    -- Show/Hide the most valuable quest reward, if required
    if ((QuestFrame:IsShown()) or
        (QuestLogFrame:IsShown()) or
        (QuestLogDetailFrame:IsShown())) then
      NigglesQuestsItemInfoUpdate();
    end

    -- Force the Archaeology pages to update themselves, if required
    if ((ArchaeologyFrameSummaryPage) and
        (ArchaeologyFrameSummaryPage:IsShown())) then
      NigglesArchaeologySummaryUpdate();
    end
    if ((ArchaeologyFrameCompletedPage) and
        (ArchaeologyFrameCompletedPage:IsShown())) then
      NigglesArchaeologyCompletedUpdate();
    end

    -- Set up the toggles in the Mount & Pet Journal
    NigglesCompanionsTogglesSetup();
  end

  return;
end

--
-- Function to enable/disable frames in the Niggles Options Warnings panel
--
function NigglesOptionsWarningsEnableFrames()
  -- Enable or disable frames based on the updated options
  if (Niggles.updatedOptions.settings.WarningCurrency == true) then
    UIDropDownMenu_EnableDropDown(NigglesOptionsWarningsWarningCurrencyMenu);
  else
    UIDropDownMenu_DisableDropDown(NigglesOptionsWarningsWarningCurrencyMenu);
  end

  return;
end

--
-- Function to register/un-register for events, based on the options
--
function NigglesRegisterForEvents()
  -- Check if new achievements should be processed
  if (NigglesDBC.settings.UiAchievementSearch) then
    Niggles:RegisterEvent("ACHIEVEMENT_EARNED");
  else
    Niggles:UnregisterEvent("ACHIEVEMENT_EARNED");
  end
  NigglesAchievementsEnable(true);

  -- Check if clicking unit buttons in the raid panel should target units
  if ((NigglesDBC.settings.RaidsTarget == true) or
      (NigglesDBC.settings.RaidsAutoPromote == true)) then
    Niggles:RegisterEvent("GROUP_ROSTER_UPDATE");
    Niggles:RegisterEvent("PLAYER_REGEN_ENABLED");
  else
    Niggles:UnregisterEvent("GROUP_ROSTER_UPDATE");
    Niggles:UnregisterEvent("PLAYER_REGEN_ENABLED");
  end
  NigglesRaidButtonsSetTargets();

  -- Check if automatic loot settings are set
  if ((NigglesDBC.settings.RaidsAutoLootMethod == true) or
      (NigglesDBC.settings.RaidsAutoLootThold == true)) then
    Niggles:RegisterEvent("PARTY_CONVERTED_TO_RAID");
  else
    Niggles:UnregisterEvent("PARTY_CONVERTED_TO_RAID");
  end

  -- Check if warnings should be generated for pets growling
  if (NigglesDBC.settings.WarningPetGrowl == true) then
    Niggles:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED");
  else
    Niggles:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED");
  end

  -- Check if warnings should be generated when entering combat
  if ((NigglesDBC.settings.WarningRod) or
      (NigglesDBC.settings.WarningLance) or
      (NigglesDBC.settings.WarningLowLevelItems) or
      (NigglesDBC.settings.WarningNoForm)) then
    Niggles:RegisterEvent("PLAYER_REGEN_DISABLED");
  else
    Niggles:UnregisterEvent("PLAYER_REGEN_DISABLED");
  end

  -- Check if warnings should be generated for unequipped pet battle items
  if (NigglesDBC.settings.WarningPetBattle) then
    Niggles:RegisterEvent("PET_BATTLE_OPENING_DONE");
  else
    Niggles:UnregisterEvent("PET_BATTLE_OPENING_DONE");
  end

  -- Check if warnings should be generated for currencies
  if (NigglesDBC.settings.WarningCurrency) then
    Niggles:RegisterEvent("CURRENCY_DISPLAY_UPDATE");
    NigglesCurrencyCheck();
  else
    Niggles:UnregisterEvent("CURRENCY_DISPLAY_UPDATE");
  end

  -- Check if updates to the guild roster should be processed
  if (NigglesDBC.settings.CalEventsOnline == true) then
    Niggles:RegisterEvent("GUILD_ROSTER_UPDATE");
  else
    Niggles:UnregisterEvent("GUILD_ROSTER_UPDATE");
  end

  -- Check if updates to calendar events should be processed
  if (NigglesDBC.settings.CalEventsOnline) then
    Niggles:RegisterEvent("CALENDAR_UPDATE_INVITE_LIST");
  else
    Niggles:UnregisterEvent("CALENDAR_UPDATE_INVITE_LIST");
  end

  -- Check if changes to the player's equipment should be monitored
  if (NigglesDBC.settings.WarningLowLevelItems) then
    Niggles:RegisterEvent("PLAYER_EQUIPMENT_CHANGED");
    NigglesEquipmentScan();
  else
    Niggles:UnregisterEvent("PLAYER_EQUIPMENT_CHANGED");
  end

  -- Enable/disable logging
  LoggingChat(NigglesDBC.settings.LogChat);
  LoggingCombat(NigglesDBC.settings.LogCombat);

  -- Enable/disable pet summoning
  NigglesPetsSetMask();
  NigglesPetsRegisterEvents();

  -- Set the font for the mail system
  if (NigglesDBC.settings.UiMailFont == true) then
    OpenMailBodyText:SetFontObject("QuestFont");
    SendMailBodyEditBox:SetFontObject("QuestFont");
  else
    OpenMailBodyText:SetFontObject("MailTextFontNormal");
    SendMailBodyEditBox:SetFontObject("MailTextFontNormal");
  end

  -- Show/Hide the Niggles Guild roster filter
  if (NigglesGuildRosterButton ~= nil) then
    if (NigglesDBC.settings.UiGuildRosterFilter == true) then
      NigglesGuildRosterButton:Show();
    else
      NigglesGuildRosterButton:Hide();
      NigglesGuildRosterShow(false);
    end
  end

  -- Enable/Disable the Map Location
  NigglesMapLocationEnable();

  -- Enable/Disable the Flights panel
  NigglesFlightsEnable();

  -- Show/Hide the frame used to provide enhancements for the Artifact icon
  if (NigglesArtifactIcon ~= nil) then
    if (NigglesDBC.settings.ProfArchEnhIcon == true) then
      NigglesArtifactIcon:Show();
    else
      NigglesArtifactIcon:Hide();
    end
  end

  -- Enable/Disable Merchant Search
  NigglesMerchantSearchEnable();

  -- Show/Hide the Merchant Split Stack 
  NigglesStackSplitOnShow();

  -- Enable/Disable the Mount & Pet Journals
  NigglesMountJournalEnable();
  NigglesPetTogglesEnable();

  -- Show/Hide the buttons for toggling class colours
  if (NigglesDBC.settings.UiChatClassColor == true) then
    NigglesChatConfigChatButton:Show();
    NigglesChatConfigChannelButton:Show();
  else
    NigglesChatConfigChatButton:Hide();
    NigglesChatConfigChannelButton:Hide();
  end

  -- Register/Unregister for guild roster updates
  if (NigglesDBC.settings.CalEventsOnline == true) then
    Niggles:RegisterEvent("GUILD_ROSTER_UPDATE");
  else
    Niggles:UnregisterEvent("GUILD_ROSTER_UPDATE");
  end

  -- Hide the 'Calendar Events' frame if not enabled
  if (NigglesDBC.settings.CalEventsList ~= true) then
    HideUIPanel(NigglesCalendarEvents);
  end

  return;
end

--
-- Function to make sure all the settings are initialised
--
function NigglesSettingsInit()
  -- Destroy any obsolete data
  if (next(NigglesDB)) then
    for key, _ in pairs(NigglesDB) do
      if (requiredDB[key] == nil) then
        NigglesDB[key] = nil;
      end
    end
  end
  if (next(NigglesDBC)) then
    for key, _ in pairs(NigglesDBC) do
      if (requiredDBC[key] == nil) then
        NigglesDBC[key] = nil;
      end
    end
  end
  if ((NigglesDBC.settings ~= nil) and
      (next(NigglesDBC.settings))) then
    for key, _ in pairs(NigglesDBC.settings) do
      if (defaultSettings[key] == nil) then
        NigglesDBC.settings[key] = nil;
      end
    end
  end
  if ((NigglesDBC.mounts ~= nil) and
      (next(NigglesDBC.mounts))) then
    for key, value in pairs(NigglesDBC.mounts) do
      if (value == 0) then
         NigglesDBC.mounts[key] = nil;
      end
    end
  end

  -- Initialise any missing required Account or Character data
  for key, _ in pairs(requiredDB) do
    if (NigglesDB[key] == nil) then
      NigglesDB[key] = {};
    end
  end
  for key, _ in pairs(requiredDBC) do
    if (NigglesDBC[key] == nil) then
      NigglesDBC[key] = {};
    end
  end

  -- Make sure all settings are initialised
  for key, value in pairs(defaultSettings) do
    if (NigglesDBC.settings[key] == nil) then
      NigglesDBC.settings[key] = value;
    end
  end

  -- Adjust settings based on class/race
  if (Niggles.playerForm == nil) then
    NigglesDBC.settings.WarningNoForm = false;
  end

  return;
end
